
from ATOM3 import ATOM3

import exporter_ClassDiagram 
import exporter_FileOutputer
import exporter_StateChartDiagram

reload(exporter_ClassDiagram)
reload(exporter_FileOutputer)
reload(exporter_StateChartDiagram)

def export(existingATOM3, model):
    
    classes = []
    classesNames = ""
                
    for classEntity in model.listNodes['CD_Class3']:
        c = exporter_ClassDiagram.ClassDiagram(classEntity, existingATOM3)
        print "Class", c.name, "has been loaded in exporter."
        classesNames = classesNames + " " + c.name
        classes.append(c)
                    
    for c in classes:
        exporter = GenericExporter(c)
        exporter.run()
    
    return classesNames

class GenericExporter:
        
    def __init__(self, classDiagram):
        
        self.classDiagram = classDiagram
        if classDiagram.stateChart == 0:
            self.stateChart = None
        else:
            self.stateChart = classDiagram.stateChart 
        
    def run(self):
        
        methodList = []
        constructorList = []
        destructorList = []
                
        #self.fOut = exporter_FileOutputer.FileOutputer("d:\\dev\\c304-statechart\\src\\" + self.classDiagram.name + ".py")
        self.fOut = exporter_FileOutputer.FileOutputer(self.classDiagram.name + ".py")
        
        for code in self.classDiagram.preClassCode:
            self.writeCode(code)
        self.fOut.write()            
        
        self.fOut.write("from time import *")
        self.fOut.write("")
        
        self.fOut.write("class " + self.classDiagram.name + ":")
        self.fOut.indent()
        self.fOut.write()
        
        for attribute, type, initValue in self.classDiagram.attributes:
            if type == "Float" :
                self.fOut.write(attribute + " = " + initValue)
            if type == "Integer" :
                self.fOut.write(attribute + " = " + initValue)
            if type == "Boolean" :
                self.fOut.write(attribute + " = " + initValue)                
            if type == "String" :
                self.fOut.write(attribute + " = " + initValue)
                    
        for name, type in self.classDiagram.objectRef:
            self.fOut.write(name + " = None")
        
        self.fOut.write()
        
        if self.stateChart != None:
            self.initCodeForChart(self.classDiagram.constructors)
            self.fOut.write("\n")
        else:
            for parameters, body in self.classDiagram.constructors:
                self.writeMethod("__init__", parameters, "", body)
        
        for name, parameters, returnType, body in self.classDiagram.methods:
            self.writeMethod(name, parameters, returnType, body)        
            
        for body in self.classDiagram.destructors:
            self.writeMethod("destructor", "", "void", body)        
            
        if self.stateChart != None:
            self.processChart()
            
            self.fOut.write("def GEN(self, target, event) :")
            self.fOut.indent()
            self.fOut.write("target.queue.append(event)")
            self.fOut.dedent()
         
            
        
        self.fOut.dedent()
        self.fOut.close()
    
    def writeCode(self, body):
        lineByLine = body.split("\n")
        for line in lineByLine:
            self.fOut.write(line)
    
    def writeConstructor(self, name, parameters, body):
        
        self.fOut.write()
        if parameters == []:
            self.fOut.write("def __init__(self):\n")
        else:
            self.fOut.write("def __init__(self," + parameters + "):\n")
        self.fOut.indent()
        self.writeCode(body)
        self.fOut.write()            
        self.fOut.write("self.current_state = \"" + self.stateChart.initState + "\"")
        self.fOut.write("self.last_state = None")
        self.fOut.write("self.current_time = clock()")
        self.fOut.write("self.queue = []")
        self.fOut.write()
        if self.stateChart.numberTimeTransitions != 0:
            self.fOut.write("self.timers = []")
            for i in range(0, self.stateChart.numberTimeTransitions):
                self.fOut.write("self.timers.append(-1)")
            for timer, exp in self.stateChart.initTimers:
                self.fOut.write("self.timers["+timer+"] = " + exp + " + self.current_time")            
            self.fOut.write()
            
        for code in self.stateChart.initCode:
            self.writeCode(code)

        self.fOut.dedent()
        self.fOut.write("")
    
    def writeMethod(self, name, parameters, returnType, body):
        
        if parameters == []:
            self.fOut.write("def " + name + "(self):\n")
        else:
            self.fOut.write("def " + name + "(self," + parameters + "):\n")
        self.fOut.indent()
        self.fOut.write()
        self.writeCode(body)
        self.fOut.write()
        self.fOut.dedent()
    
    def initCodeForChart(self, constructors):
                
        #writeConstructor(self.fOut, self.classDiagram.name, "", "")
        if constructors != []:
            for parameters, body in constructors:
                self.writeConstructor(self.classDiagram.name, parameters, body)    
        else:
            self.writeConstructor("__init__", "", "")    
    
    def processChart(self):
        
        self.fOut.write("def process(self):\n")
        self.fOut.indent()
        self.fOut.write("self.current_time = clock()\n")
        if self.stateChart.numberTimeTransitions != 0:
            self.fOut.write("# Lets check the timers")
            self.fOut.write("for i in range(len(self.timers)):")            
            self.fOut.indent()
            self.fOut.write("if self.timers[i] != -1 and self.timers[i] > self.current_time:")
            self.fOut.indent()
            self.fOut.write("self.queue.append(\"_\" + str(i) + \"after\")")            
            self.fOut.dedent()
            self.fOut.write("")
            self.fOut.dedent()
            self.fOut.write("")
    
        self.fOut.write()        
        self.fOut.write("while len(self.queue) != 0:")
        self.fOut.indent()
        self.fOut.write(" ")
        self.fOut.write("event = self.queue[0]")
        self.fOut.write("del self.queue[0]")
        self.fOut.write(" ")
        for state in self.stateChart.basics:
            # If the state has transition
            if self.stateChart.transitionData[state] != [] :            
                self.fOut.write("if  self.current_state == \"" + state.name.getValue() + "\":\n")
                self.fOut.indent()
                self.processTransition(state)
                self.fOut.write("event = None")
                self.fOut.write("")
                self.fOut.dedent()
                self.fOut.write("")
        
        self.fOut.dedent()
        self.fOut.write("")
            
        self.fOut.dedent()
        self.fOut.write("\n")
        
    def processTransition(self, state):
        
        transitions = self.stateChart.transitionData[state]
        
        for transition in transitions:
                                        
            if "event" in transition:
                condition = "event == \"" + transition["event"] + "\""
            else:
                condition = ""
            
            if "guard" in transition:
                if transition["guard"] == "1":
                    transition["guard"] = "True"
                        
                if condition == "":
                    condition = transition["guard"]
                else :
                    condition += " and " + transition["guard"]
            
            self.fOut.write("if " + condition + ":")
            self.fOut.indent()
            
            for key, value in transition["timersToReset"]:
                self.fOut.write("self.timers [" + key + "] = -1;")
                
            for key, value in transition["historyToSet"]:
                self.fOut.write("self.history [" + key + "] = " + value)
                
            for key, value in transition["timersToSet"]:
                self.fOut.write("self.timers [" + key + "] = " + value + "+ self.current_time")
            
            for code in transition["exitCode"]:
                self.writeCode(code)
                
            for code in transition["triggerCode"]:
                self.writeCode(code)
            
            for code in transition["enterCode"]:
                self.writeCode(code)
            
            
            self.fOut.write("self.current_state = \"" + transition["destinationState"] + "\"")
            
            self.fOut.dedent()
            self.fOut.write()
        